
/*
 * pbrt source code Copyright(c) 1998-2005 Matt Pharr and Greg Humphreys
 *
 * All Rights Reserved.
 * For educational use only; commercial use expressly forbidden.
 * NO WARRANTY, express or implied, for this software.
 * (See file License.txt for complete license)
 */

// uber.cpp*
#include "pbrt.h"
#include "material.h"



// UberMaterial Class Declarations
class UberMaterial : public Material {
public:
	// UberMaterial Method Declarations
	UberMaterial(Reference<Texture<Spectrum> > kd,
		Reference<Texture<Spectrum> > ks,
		Reference<Texture<Spectrum> > kr,
		Reference<Texture<float> > rough,
		Reference<Texture<Spectrum> > op,
		Reference<Texture<float> > bump, 
		float G,
		float SigmaS,
		float SigmaA,
		float N) {
		Kd = kd;
		Ks = ks;
		Kr = kr;
		roughness = rough;
		opacity = op;
		bumpMap = bump;

		//used for subsurface scattering
		g = G;
		sigmas = SigmaS;
		sigmaa = SigmaA;
		if(g > 0.7f) {
			sigmasnot = (1.0f - g) * sigmas;
		}
		else
			sigmasnot = sigmas;
		sigmatnot = sigmasnot + sigmaa;
		lu = 1.0f / sigmatnot;
		n = N;
		Fdr = -1.440 / (n*n) + 0.710 / n + 0.668 + 0.0636*n;
		Fdt = 1.f - Fdr;
		A = (1.f + Fdr) / Fdt;
		Zv = lu * (1.f + (4.f/3.f)*A);
		Zr = lu;
		sigmatr = sqrtf(3.f*sigmaa*sigmatnot);
	}
	BSDF *GetBSDF(const DifferentialGeometry &dgGeom, const DifferentialGeometry &dgShading) const;
	//bool isSubsurface() { return true; }
	//void StoreSamples(Reference<Shape> s);
	virtual float getMaterialN() const {return n;}
	virtual SSMaterialData getMaterialData() const {
		SSMaterialData ssmd;
		ssmd.G_Fdr = Fdr;
		ssmd.G_Fdt = Fdt;
		ssmd.G_Sigmatr = sigmatr;
		ssmd.G_Zr = Zr;
		ssmd.G_Zv = Zv;
		return ssmd;
	}
private:
	// UberMaterial Private Data
	Reference<Texture<Spectrum> > Kd, Ks, Kr, opacity;
	Reference<Texture<float> > roughness;
	Reference<Texture<float> > bumpMap;
	//used for subsurface scattering
	float g; //mean cosine of scattering angle
	float sigmas; //scattering coef.
	float sigmasnot; //reduced scattering coef.
	float sigmaa; //absorption coef. 
	float sigmatnot; //reduced extinction coef.
	float sigmatr;
	float lu; //mean-free path
	float n; //index of refraction
	float Fdr;
	float Fdt;
	float A;
	float Zr, Zv;
	//int nSamples;
	//SamplePoint *mySamplePts;
	//Point *samplePoints;
	//float *sampleArea;
	//Spectrum *sampleIrradiance;
};
// UberMaterial Method Definitions
BSDF *UberMaterial::GetBSDF(const DifferentialGeometry &dgGeom, const DifferentialGeometry &dgShading) const {
	// Allocate _BSDF_, possibly doing bump-mapping with _bumpMap_
	DifferentialGeometry dgs;
	if (bumpMap)
		Bump(bumpMap, dgGeom, dgShading, &dgs);
	else
		dgs = dgShading;
	BSDF *bsdf = BSDF_ALLOC(BSDF)(dgs, dgGeom.nn);

	Spectrum op = opacity->Evaluate(dgs).Clamp();
	if (op != Spectrum(1.)) {
	    BxDF *tr = BSDF_ALLOC(SpecularTransmission)(-op + Spectrum(1.), 1., 1.);
	    bsdf->Add(tr);
	}

	Spectrum kd = op * Kd->Evaluate(dgs).Clamp();
	if (!kd.Black()) {
	    BxDF *diff = BSDF_ALLOC(Lambertian)(kd);
	    bsdf->Add(diff);
	}

	Spectrum ks = op * Ks->Evaluate(dgs).Clamp();
	if (!ks.Black()) {
	    Fresnel *fresnel = BSDF_ALLOC(FresnelDielectric)(1.5f, 1.f);
	    float rough = roughness->Evaluate(dgs);
	    BxDF *spec = BSDF_ALLOC(Microfacet)(ks, fresnel, BSDF_ALLOC(Blinn)(1.f / rough));
	    bsdf->Add(spec);
	}

	Spectrum kr = op * Kr->Evaluate(dgs).Clamp();
	if (!kr.Black()) {
		Fresnel *fresnel = BSDF_ALLOC(FresnelDielectric)(1.5f, 1.f);
		bsdf->Add(BSDF_ALLOC(SpecularReflection)(kr, fresnel));
	}

	return bsdf;
}

//for subsurface scattering
//void UberMaterial::StoreSamples(Reference<Shape> s) {
//	s->GetSamples(&nSamples, mySamplePts);
//}

// UberMaterial Dynamic Creation Routine
extern "C" DLLEXPORT Material * CreateMaterial(const Transform &xform,
		const TextureParams &mp) {
	Reference<Texture<Spectrum> > Kd = mp.GetSpectrumTexture("Kd", Spectrum(1.f));
	Reference<Texture<Spectrum> > Ks = mp.GetSpectrumTexture("Ks", Spectrum(1.f));
	Reference<Texture<Spectrum> > Kr = mp.GetSpectrumTexture("Kr", Spectrum(0.f));
	Reference<Texture<float> > roughness = mp.GetFloatTexture("roughness", .1f);
	Reference<Texture<Spectrum> > opacity = mp.GetSpectrumTexture("opacity", 1.f);
	Reference<Texture<float> > bumpMap = mp.GetFloatTexture("bumpmap", 0.f);
	//used for subsurface scattering - default values for skin
	float G = mp.FindFloat("g", 0.79f);
	float SigmaS = mp.FindFloat("sigmas", 50.0f);
	float SigmaA = mp.FindFloat("sigmaa", 3.8f);
	float N = mp.FindFloat("n", 1.4f); 
	return new UberMaterial(Kd, Ks, Kr, roughness, opacity, bumpMap, G, SigmaS, SigmaA, N);
}
